#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['asyncio_backport']

package_data = \
{'': ['*']}

extras_require = \
{'test': ['pytest', 'coverage_conditional_plugin', 'pytest-cov', 'mypy']}

setup(name='asyncio-backport',
      version='0.1.1',
      description='A backport of asyncio features from newer versions',
      author=None,
      author_email='Thomas Grainger <asyncio-backport@graingert.co.uk>',
      url=None,
      packages=packages,
      package_data=package_data,
      extras_require=extras_require,
     )
